
import json
import aiohttpx
from typing import Any, Optional, Union, Dict
from lazyops.types import BaseModel, lazyproperty

__all__ = [
    "OpenAIError",
    "ExceptionModel",
    "fatal_exception",
    "APIError",
    "TryAgain",
    "APIConnectionError",
    "Timeout",
    "InvalidRequestError",
    "AuthenticationError",
    "PermissionError",
    "RateLimitError",
    "ServiceUnavailableError",
    "InvalidAPIType",
    "error_handler",
]

class ExceptionModel(BaseModel):
    response: aiohttpx.Response
    data: Optional[Union[Dict, Any]]
    message: Optional[str] = None
    should_retry: Optional[bool] = False

    @lazyproperty
    def headers(self):
        return self.response.headers
    
    @lazyproperty
    def stream(self) -> bool:
        return "text/event-stream" in self.headers.get("content-type", "")

    @lazyproperty
    def response_data(self):
        return self.data or self.response.json()
    
    @lazyproperty
    def http_body(self):
        body = self.data if self.stream else self.response.content
        try:
            if hasattr(body, "decode"):
                body = body.decode("utf-8")
            return body
        except (json.JSONDecodeError, UnicodeDecodeError) as e:
            raise ValueError(
                f"HTTP code {self.status_code} from API ({body})"
            ) from e
    
    @lazyproperty
    def response_json(self):
        try:
            return json.loads(self.http_body)
        except json.JSONDecodeError:
            return {}

    @lazyproperty
    def response_text(self):
        return self.response.text
    
    @lazyproperty
    def status_code(self):
        return self.response.status_code
    
    @lazyproperty
    def error_data(self) -> Dict:
        return self.response_json.get("error", {})
    
    @lazyproperty
    def request_id(self) -> str:
        return self.headers.get("request-id", None)
    
    @lazyproperty
    def organization(self) -> str:
        return self.headers.get("openai-organization", None)
    
    @lazyproperty
    def error_message(self) -> str:
        msg = self.message or ("(Error occurred while streaming.)" if self.stream else "")
        msg += " " + self.error_data.get("message", "")
        if self.error_data.get("internal_message"):
            msg += "\n\n" + self.error_data["internal_message"]
        return msg


class OpenAIError(Exception):
    def __init__(
        self,
        response: aiohttpx.Response,
        data: Optional[Union[Dict, Any]],
        message: Optional[str] = None,
        should_retry: Optional[bool] = False,
        **kwargs
    ):
        self.status = response.status_code
        self.response = response
        self.message = message
        self.exc = ExceptionModel(
            response=response,
            message=message,
            data=data,
            should_retry=should_retry,
            **kwargs
        )
    
    def __str__(self):
        msg = self.exc.error_message or "<empty message>"
        if self.exc.request_id is not None:
            return f"Request {self.exc.request_id}: {msg}"
        else:
            return msg
        
    @property
    def user_message(self):
        return self.exc.error_message
    
    def __repr__(self):
        return f"[OpenAI] {self.__class__.__name__} \
            (message={self.exc.error_message}, \
                http_status={self.exc.status_code}, \
                    request_id={self.exc.request_id})"



class APIError(OpenAIError):
    pass


class TryAgain(OpenAIError):
    pass


class Timeout(OpenAIError):
    pass


class APIConnectionError(OpenAIError):
    pass


class InvalidRequestError(OpenAIError):
    pass


class AuthenticationError(OpenAIError):
    pass


class PermissionError(OpenAIError):
    pass


class RateLimitError(OpenAIError):
    pass


class ServiceUnavailableError(OpenAIError):
    pass


class InvalidAPIType(OpenAIError):
    pass


def fatal_exception(exc):
    if isinstance(exc, OpenAIError):
        # retry on server errors and client errors
        # with 429 status code (rate limited),
        # don't retry on other client errors
        return (400 <= exc.status < 500) and exc.status != 429
    else:
        # retry on all other errors (eg. network)
        return False


def error_handler(
    response: aiohttpx.Response,
    data: Optional[Any] = None, # Line specific for streaming responses
    should_retry: Optional[bool] = False,
    **kwargs
):

    if response.status_code == 503:
        raise ServiceUnavailableError(
            response = response,
            message = "The server is overloaded or not ready yet.",
            data = data,
            should_retry = should_retry,
            **kwargs
        )
    if response.status_code == 429:
        raise RateLimitError(
            response = response,
            data = data,
            should_retry = should_retry,
            **kwargs
        )
    if response.status_code in [400, 404, 415]:
        return InvalidRequestError(
            response = response,
            data = data,
            should_retry = should_retry,
            **kwargs
        )
    if response.status_code == 401:
        return AuthenticationError(
            response = response,
            data = data,
            should_retry = should_retry,
            **kwargs
        )
    if response.status_code == 403:
        return PermissionError(
            response = response,
            data = data,
            should_retry = should_retry,
            **kwargs
        )
    if response.status_code == 409:
        return TryAgain(
            response = response,
            data = data,
            should_retry = should_retry,
            **kwargs
        )
    
    raise APIError(
        response = response,
        data = data,
        should_retry = should_retry,
        **kwargs
    )






from typing import Optional, Type, Any, Union, List, Dict
from lazyops.types import validator, lazyproperty

from async_openai.types.options import OpenAIModel, OpenAIModelType
from async_openai.types.resources import BaseResource
from async_openai.types.responses import BaseResponse
from async_openai.types.routes import BaseRoute


__all__ = [
    'EditChoice',
    'EditObject',
    'EditResponse',
    'EditRoute',
]


class EditChoice(BaseResource):
    text: str
    index: int
    logprobs: Optional[Any]
    finish_reason: Optional[str]

class EditObject(BaseResource):
    model: Optional[Union[str, OpenAIModel, Any]] = "curie"
    instruction: Optional[str]
    input: Optional[str] = ""
    n: Optional[int] = 1
    temperature: Optional[float] = 1.0
    top_p: Optional[float] = 1.0
    user: Optional[str] = None

    @validator('model', pre=True, always=True)
    def validate_model(cls, v) -> OpenAIModel:
        """
        Validate the model
        """
        if isinstance(v, OpenAIModel):
            return v
        if isinstance(v, dict):
            return OpenAIModel(**v)
        return OpenAIModel(value = v, mode = 'edit')


class EditResponse(BaseResponse):
    choices: Optional[List[EditChoice]]
    choice_model: Optional[Type[BaseResource]] = EditChoice


    @lazyproperty
    def text(self) -> str:
        """
        Returns the text for the edits
        """
        if self.choices:
            return ''.join([choice.text for choice in self.choices])
        return self._response.text


class EditRoute(BaseRoute):
    input_model: Optional[Type[BaseResource]] = EditObject
    response_model: Optional[Type[BaseResource]] = EditResponse

    @lazyproperty
    def api_resource(self):
        return 'edits'

    def create(
        self, 
        input_object: Optional[Type[BaseResource]] = None,
        **kwargs
    ) -> EditResponse:
        """
        
        """
        return super().create(input_object = input_object, **kwargs)
    
    async def async_create(
        self, 
        input_object: Optional[Type[BaseResource]] = None,
        **kwargs
    ) -> EditResponse:
        """
        
        """
        return await super().async_create(input_object = input_object, **kwargs)

    






import aiohttpx
import json

from typing import Optional, Type, Any, Union, List, Dict, Iterator
from lazyops.types import validator, lazyproperty

from async_openai.types.options import OpenAIModel, OpenAIModelType
from async_openai.types.resources import BaseResource
from async_openai.types.responses import BaseResponse
from async_openai.types.routes import BaseRoute
from async_openai.utils import logger


__all__ = [
    'CompletionChoice',
    'CompletionObject',
    'CompletionResponse',
    'CompletionRoute',
]


class CompletionChoice(BaseResource):
    text: str
    index: int
    logprobs: Optional[Any]
    finish_reason: Optional[str]


class CompletionObject(BaseResource):
    model: Optional[Union[OpenAIModel, str, Any]] = "davinci"
    prompt: Optional[Union[str, List[str]]] = '<|endoftext|>'
    suffix: Optional[str] = None
    max_tokens: Optional[int] = 16
    temperature: Optional[float] = 1.0
    top_p: Optional[float] = 1.0
    n: Optional[int] = 1
    stream: Optional[bool] = False
    logprobs: Optional[int] = None
    echo: Optional[bool] = False
    stop: Optional[Union[str, List[str]]] = None
    presence_penalty: Optional[float] = 0.0
    frequency_penalty: Optional[float] = 0.0
    best_of: Optional[int] = 1
    logit_bias: Optional[Dict[str, float]] = None
    user: Optional[str] = None

    @validator('model', pre=True, always=True)
    def validate_model(cls, v, values: Dict[str, Any]) -> OpenAIModel:
        """
        Validate the model
        """
        if not v and values.get('engine'):
            v = values.get('engine')
        if isinstance(v, OpenAIModel):
            return v
        if isinstance(v, dict):
            return OpenAIModel(**v)
        return OpenAIModel(value = v, mode = 'completion')

    @validator('max_tokens')
    def validate_max_tokens(cls, v: int) -> int:
        """
        Max tokens is 4096 / 8192
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-max-tokens
        """
        return None if v is None else max(0, min(v, 8192))
    
    @validator('temperature')
    def validate_temperature(cls, v: float) -> float:
        """
        Min Temperature is 0.0
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-temperature
        """
        return None if v is None else max(0.0, v)
    
    @validator('top_p')
    def validate_top_p(cls, v: float) -> float:
        """
        Min Top Probability is 0.0
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-top-probability
        """
        return None if v is None else max(0.0, v)

    @validator('logprobs')
    def validate_logprobs(cls, v: int) -> int:
        """
        Max logprobs is 5
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-logprobs
        """
        return None if v is None else max(0, min(v, 5))
    
    @validator('presence_penalty')
    def validate_presence_penalty(cls, v: float) -> float:
        """
        Min Presence Penalty is -2.0, Max is 2.0
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-presence-penalty
        """
        return None if v is None else max(0.0, min(v, 2.0))
    
    @validator('frequency_penalty')
    def validate_frequency_penalty(cls, v: float) -> float:
        """
        Min Frequency Penalty is -2.0, Max is 2.0
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-frequency-penalty
        """
        return None if v is None else max(0.0, min(v, 2.0))
    

    def dict(self, *args, exclude: Any = None, **kwargs):
        """
        Returns the dict representation of the response
        """
        data = super().dict(*args, exclude = exclude, **kwargs)
        if data.get('model'):
            data['model'] = data['model'].value
        return data


    

class CompletionResponse(BaseResponse):
    choices: Optional[List[CompletionChoice]]
    choice_model: Optional[Type[BaseResource]] = CompletionChoice

    @lazyproperty
    def text(self) -> str:
        """
        Returns the text for the completions
        """
        if self.choices_results:
            return ''.join([choice.text for choice in self.choices])
        return self._response.text
    
    @lazyproperty
    def completion_model(self):
        """
        Returns the model for the completions
        """
        return OpenAIModel(value=self.model, mode='completion') if self.model else None
    
    @lazyproperty
    def consumption(self) -> int:
        """
        Returns the consumption for the completions
        """ 
        if self.usage and self.completion_model:
            return self.completion_model.get_cost(total_tokens = self.usage.total_tokens)
        return None


    def dict(self, *args, exclude: Any = None, **kwargs):
        """
        Returns the dict representation of the response
        """
        data = super().dict(*args, exclude = exclude, **kwargs)
        if data.get('completion_model'):
            data['completion_model'] = data['completion_model'].dict()
        return data


    def handle_stream(
        self,
        response: aiohttpx.Response
    ) -> Iterator[Dict]:
        texts = {}
        for line in response.iter_lines():
            if not line: continue
            if "data: [DONE]" in line:
                continue
            if line.startswith("data: "):
                line = line[len("data: ") :]
            if not line.strip(): continue
            try:

                item = json.loads(line)
                self.handle_stream_metadata(item)
                for n, choice in enumerate(item['choices']):
                    if not texts.get(n):
                        texts[n] = {
                            'index': choice['index'],
                            'text': choice['text'],
                        }
                        self.usage.completion_tokens += 1
                    elif choice['finish_reason'] != 'stop':
                        texts[n]['text'] += choice['text']
                        self.usage.completion_tokens += 1

                    else:
                        texts[n]['finish_reason'] = choice['finish_reason']
                        yield texts.pop(n)

            except Exception as e:
                logger.error(f'Error: {line}: {e}')

        yield from texts.values()
        self.usage.total_tokens = self.usage.completion_tokens


class CompletionRoute(BaseRoute):
    input_model: Optional[Type[BaseResource]] = CompletionObject
    response_model: Optional[Type[BaseResource]] = CompletionResponse

    @lazyproperty
    def api_resource(self):
        return 'completions'

    @lazyproperty
    def root_name(self):
        return 'completion'
    
    def create(
        self, 
        input_object: Optional[Type[BaseResource]] = None,
        **kwargs
    ) -> CompletionResponse:
        """
        Creates a completion for the provided prompt and parameters

        Usage:

        ```python
        >>> result = OpenAI.completions.create(
        >>>    prompt = 'say this is a test',
        >>>    max_tokens = 4,
        >>>    stream = True
        >>> )
        ```

        **Parameters:**

        :model (required): ID of the model to use. You can use the List models API 
        to see all of your available models,  or see our Model overview for descriptions of them.
        Default: `text-curie-001`
        
        :prompt (optional): The prompt(s) to generate completions for, encoded as 
        a string, array of strings, array of tokens, or array of token arrays.
        Note that `<|endoftext|>` is the document separator that the model sees during 
        training, so if a prompt is not specified the model will generate as if from 
        the beginning of a new document.
        Default: `<|endoftext|>`

        :suffix (optional): The suffix that comes after a completion of inserted text.
        Default: `None`

        :max_tokens (optional): The maximum number of tokens to generate in the completion.
        The token count of your prompt plus `max_tokens` cannot exceed the model's context length. 
        Most models have a context length of 2048 tokens (except for the newest models, which 
        support 4096).
        Default: `16`

        :temperature (optional): What sampling temperature to use. Higher values means 
        the model will take more risks. Try 0.9 for more creative applications, and 0 (argmax sampling) 
        for ones with a well-defined answer. We generally recommend altering this or `top_p` but not both.
        Default: `1.0`

        :top_p (optional): An alternative to sampling with `temperature`, called nucleus 
        sampling, where the model considers the results of the tokens with `top_p` probability mass. 
        So `0.1` means only  the tokens comprising the top 10% probability mass are considered.
        We generally recommend altering this or `temperature` but not both
        Default: `1.0`

        :n (optional): How many completions to generate for each prompt.
        Note: Because this parameter generates many completions, it can quickly 
        consume your token quota. Use carefully and ensure that you have reasonable 
        settings for `max_tokens` and stop.
        Default: `1`

        :stream (optional): Whether to stream back partial progress. 
        If set, tokens will be sent as data-only server-sent events as they become 
        available, with the stream terminated by a `data: [DONE]` message. This is 
        handled automatically by the Client and enables faster response processing.
        Default: `False`

        :logprobs (optional): Include the log probabilities on the `logprobs` 
        most likely tokens, as well the chosen tokens. For example, if `logprobs` is 5, 
        the API will return a list of the 5 most likely tokens. The API will always 
        return the logprob of the sampled token, so there may be up to `logprobs+1` 
        elements in the response. The maximum value for `logprobs` is 5.
        Default: `None`

        :echo (optional): Echo back the prompt in addition to the completion
        Default: `False`

        :stop (optional): Up to 4 sequences where the API will stop generating 
        further tokens. The returned text will not contain the stop sequence.
        Default: `None`

        :presence_penalty (optional): Number between `-2.0` and `2.0`. Positive values 
        penalize new tokens based on whether they appear in the text so far, increasing the 
        model's likelihood to talk about new topics
        Default: `0.0`

        :frequency_penalty (optional): Number between `-2.0` and `2.0`. Positive 
        values penalize new tokens based on their existing frequency in the text so 
        far, decreasing the model's likelihood to repeat the same line verbatim.
        Default: `0.0`

        :best_of (optional): Generates `best_of` completions server-side and returns 
        the "best" (the one with the highest log probability per token). Results cannot be streamed.
        When used with `n`, `best_of` controls the number of candidate completions and n specifies how 
        many to return – `best_of` must be greater than `n`.
        Note: Because this parameter generates many completions, it can quickly consume your token quota. 
        Use carefully and ensure that you have reasonable settings for `max_tokens` and `stop`.
        Default: `1`

        :logit_bias (optional): Modify the likelihood of specified tokens appearing in the completion.
        Accepts a json object that maps tokens (specified by their token ID in the GPT tokenizer) to an associated 
        bias value from -100 to 100. You can use this tokenizer tool (which works for both GPT-2 and GPT-3) to 
        convert text to token IDs. Mathematically, the bias is added to the logits generated by the model prior 
        to sampling. The exact effect will vary per model, but values between -1 and 1 should decrease or increase 
        likelihood of selection; values like -100 or 100 should result in a ban or exclusive selection of the 
        relevant token.
        As an example, you can pass `{"50256": -100}` to prevent the `<|endoftext|>` token from being generated.
        Default: `None`

        :user (optional): A unique identifier representing your end-user, which can help OpenAI to 
        monitor and detect abuse.
        Default: `None`

        Returns: `CompletionResult`
        """
        return super().create(input_object=input_object, **kwargs)

    async def async_create(
        self, 
        input_object: Optional[Type[BaseResource]] = None,
        **kwargs
    ) -> CompletionResponse:
        """
        Creates a completion for the provided prompt and parameters

        Usage:

        ```python
        >>> result = await OpenAI.completions.async_create(
        >>>    prompt = 'say this is a test',
        >>>    max_tokens = 4,
        >>>    stream = True
        >>> )
        ```

        **Parameters:**

        :model (required): ID of the model to use. You can use the List models API 
        to see all of your available models,  or see our Model overview for descriptions of them.
        Default: `text-curie-001`
        
        :prompt (optional): The prompt(s) to generate completions for, encoded as 
        a string, array of strings, array of tokens, or array of token arrays.
        Note that `<|endoftext|>` is the document separator that the model sees during 
        training, so if a prompt is not specified the model will generate as if from 
        the beginning of a new document.
        Default: `<|endoftext|>`

        :suffix (optional): The suffix that comes after a completion of inserted text.
        Default: `None`

        :max_tokens (optional): The maximum number of tokens to generate in the completion.
        The token count of your prompt plus `max_tokens` cannot exceed the model's context length. 
        Most models have a context length of 2048 tokens (except for the newest models, which 
        support 4096).
        Default: `16`

        :temperature (optional): What sampling temperature to use. Higher values means 
        the model will take more risks. Try 0.9 for more creative applications, and 0 (argmax sampling) 
        for ones with a well-defined answer. We generally recommend altering this or `top_p` but not both.
        Default: `1.0`

        :top_p (optional): An alternative to sampling with `temperature`, called nucleus 
        sampling, where the model considers the results of the tokens with `top_p` probability mass. 
        So `0.1` means only  the tokens comprising the top 10% probability mass are considered.
        We generally recommend altering this or `temperature` but not both
        Default: `1.0`

        :n (optional): How many completions to generate for each prompt.
        Note: Because this parameter generates many completions, it can quickly 
        consume your token quota. Use carefully and ensure that you have reasonable 
        settings for `max_tokens` and stop.
        Default: `1`

        :stream (optional): Whether to stream back partial progress. 
        If set, tokens will be sent as data-only server-sent events as they become 
        available, with the stream terminated by a `data: [DONE]` message. This is 
        handled automatically by the Client and enables faster response processing.
        Default: `False`

        :logprobs (optional): Include the log probabilities on the `logprobs` 
        most likely tokens, as well the chosen tokens. For example, if `logprobs` is 5, 
        the API will return a list of the 5 most likely tokens. The API will always 
        return the logprob of the sampled token, so there may be up to `logprobs+1` 
        elements in the response. The maximum value for `logprobs` is 5.
        Default: `None`

        :echo (optional): Echo back the prompt in addition to the completion
        Default: `False`

        :stop (optional): Up to 4 sequences where the API will stop generating 
        further tokens. The returned text will not contain the stop sequence.
        Default: `None`

        :presence_penalty (optional): Number between `-2.0` and `2.0`. Positive values 
        penalize new tokens based on whether they appear in the text so far, increasing the 
        model's likelihood to talk about new topics
        Default: `0.0`

        :frequency_penalty (optional): Number between `-2.0` and `2.0`. Positive 
        values penalize new tokens based on their existing frequency in the text so 
        far, decreasing the model's likelihood to repeat the same line verbatim.
        Default: `0.0`

        :best_of (optional): Generates `best_of` completions server-side and returns 
        the "best" (the one with the highest log probability per token). Results cannot be streamed.
        When used with `n`, `best_of` controls the number of candidate completions and n specifies how 
        many to return – `best_of` must be greater than `n`.
        Note: Because this parameter generates many completions, it can quickly consume your token quota. 
        Use carefully and ensure that you have reasonable settings for `max_tokens` and `stop`.
        Default: `1`

        :logit_bias (optional): Modify the likelihood of specified tokens appearing in the completion.
        Accepts a json object that maps tokens (specified by their token ID in the GPT tokenizer) to an associated 
        bias value from -100 to 100. You can use this tokenizer tool (which works for both GPT-2 and GPT-3) to 
        convert text to token IDs. Mathematically, the bias is added to the logits generated by the model prior 
        to sampling. The exact effect will vary per model, but values between -1 and 1 should decrease or increase 
        likelihood of selection; values like -100 or 100 should result in a ban or exclusive selection of the 
        relevant token.
        As an example, you can pass `{"50256": -100}` to prevent the `<|endoftext|>` token from being generated.
        Default: `None`

        :user (optional): A unique identifier representing your end-user, which can help OpenAI to 
        monitor and detect abuse.
        Default: `None`

        Returns: `CompletionResult`
        """
        return await super().async_create(input_object=input_object, **kwargs)


import aiohttpx
import pathlib
from typing import Optional, Callable, Dict, Union, List

from lazyops.types import lazyproperty
from async_openai.schemas import *
from async_openai.types.options import ApiType
from async_openai.utils.logs import logger
from async_openai.utils.config import settings
from async_openai.routes import ApiRoutes
from async_openai.meta import OpenAIMetaClass


class OpenAIClient:
    """
    Main Client for all the routes in the API.
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        url: Optional[str] = None,
        scheme: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        api_path: Optional[str] = None,
        api_type: Optional[ApiType] = None,
        api_version: Optional[str] = None,
        organization: Optional[str] = None,
        proxies: Optional[Union[str, Dict]] = None,
        app_info: Optional[Dict[str, str]] = None,
        
        headers: Optional[Dict] = None,
        debug_enabled: Optional[bool] = None,
        on_error: Optional[Callable] = None,
        timeout: Optional[int] = None,
        max_retries: Optional[int] = None,
        ignore_errors: Optional[bool] = None,

        **kwargs
    ):
        self.api_key = api_key if api_key is not None else settings.api_key
        self.api_type = api_type if api_type is not None else settings.api_type
        self.api_version = api_version if api_version is not None else settings.api_version
        self.organization = organization if organization is not None else settings.organization
        self.proxies = proxies if proxies is not None else settings.proxies
        self.app_info = app_info if app_info is not None else settings.app_info

        self.api_url = settings.get_api_url(host = host, port = port, scheme = scheme, url = url)
        self.base_url = settings.get_base_api_url(host = host, port = port, scheme = scheme, url = url, api_path = api_path)
        self.debug_enabled = debug_enabled if debug_enabled is not None else settings.debug_enabled
        
        self.timeout = timeout if timeout is not None else settings.timeout
        self.headers = headers if headers is not None else settings.get_headers(api_key = self.api_key, api_version = self.api_version, api_type = self.api_type, organization = self.organization, app_info = self.app_info)

        self.on_error = on_error
        self.ignore_errors = ignore_errors if ignore_errors is not None else settings.ignore_errors
        self.max_retries = max_retries if max_retries is not None else settings.max_retries

        self._kwargs = kwargs
        self.log_method = logger.info if self.debug_enabled else logger.debug
        self.client = aiohttpx.Client(
            base_url = self.base_url,
            timeout = self.timeout,
        )

        self.routes = ApiRoutes(
            client = self.client,
            headers = self.headers,
            debug_enabled = self.debug_enabled,
            on_error = self.on_error,
            ignore_errors = self.ignore_errors,
            timeout = self.timeout,
            max_retries = self.max_retries,
            **self._kwargs
        )
        logger.info(f"OpenAI Client initialized: {self.client.base_url}")
        if self.debug_enabled:
            logger.debug(f"Debug Enabled: {self.debug_enabled}")


    @lazyproperty
    def completions(self) -> CompletionRoute:
        """
        Returns the `CompletionRoute` class for interacting with `Completions`.
        
        Doc: `https://beta.openai.com/docs/api-reference/completions`
        """
        return self.routes.completions
    
    @lazyproperty
    def chat(self) -> ChatRoute:
        """
        Returns the `ChatRoute` class for interacting with `Chat` components

        Doc: `https://platform.openai.com/docs/api-reference/chat`
        """
        return self.routes.chat

    @lazyproperty
    def edits(self) -> EditRoute:
        """
        Returns the `EditRoute` class for interacting with `Edits`.
        
        Doc: `https://beta.openai.com/docs/api-reference/edits`
        """
        return self.routes.edits
    
    @lazyproperty
    def embeddings(self) -> EmbeddingRoute:
        """
        Returns the `EmbeddingRoute` class for interacting with `Embeddings`.
        
        Doc: `https://beta.openai.com/docs/api-reference/embeddings`
        """
        return self.routes.embeddings
    
    @lazyproperty
    def images(self) -> ImageRoute:
        """
        Returns the `ImageRoute` class for interacting with `Images`.
        
        Doc: `https://beta.openai.com/docs/api-reference/images`
        """
        return self.routes.images
    
    @lazyproperty
    def models(self) -> ModelRoute:
        """
        Returns the `ModelRoute` class for interacting with `models`.
        
        Doc: `https://beta.openai.com/docs/api-reference/models`
        """
        return self.routes.models

    """
    Context Managers
    """

    async def async_close(self):
        await self.client.aclose()
    
    def close(self):
        self.client.close()
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
    
    async def __aenter__(self):
        return self
    
    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.async_close()


class OpenAI(metaclass = OpenAIMetaClass):
    """
    Interface for OpenAI
    """
    pass






import os
from lazyops.utils.logs import get_logger, change_logger_level

# to prevent recursive imports, we'll just use os environ here
if os.getenv('DEBUG_ENABLED') == 'True':
    logger_level = 'DEV'
else:
    logger_level: str = os.getenv('LOGGER_LEVEL', 'INFO').upper()

logger = get_logger(logger_level)


from typing import Optional, Type, Any, Union, List
from lazyops.types import validator, lazyproperty

from async_openai.types.options import OpenAIModel, OpenAIModelType
from async_openai.types.resources import BaseResource
from async_openai.types.responses import BaseResponse
from async_openai.types.routes import BaseRoute


__all__ = [
    'EmbeddingData',
    'EmbeddingObject',
    'EmbeddingResponse',
    'EmbeddingRoute',
]


class EmbeddingData(BaseResource):
    object: Optional[str] = 'embedding'
    embedding: Optional[List[float]] = []
    index: Optional[int] = 0

class EmbeddingObject(BaseResource):
    model: Optional[Union[str, OpenAIModel, Any]] = "curie"
    input: Optional[Union[List[Any], Any]]
    user: Optional[str] = None

    @validator('model', pre=True, always=True)
    def validate_model(cls, v) -> OpenAIModel:
        """
        Validate the model
        """
        if isinstance(v, OpenAIModel):
            return v
        if isinstance(v, dict):
            return OpenAIModel(**v)
        return OpenAIModel(value = v, mode = 'embedding')


class EmbeddingResponse(BaseResponse):
    data: Optional[List[EmbeddingData]]
    data_model: Optional[Type[BaseResource]] = EmbeddingData

    @lazyproperty
    def embeddings(self) -> List:
        """
        Returns the text for the response
        object
        """
        if self.data:
            return [data.embedding for data in self.data] if self.data else []
        return None


class EmbeddingRoute(BaseRoute):
    input_model: Optional[Type[BaseResource]] = EmbeddingObject
    response_model: Optional[Type[BaseResource]] = EmbeddingResponse

    @lazyproperty
    def api_resource(self):
        return 'embeddings'





import json
import pathlib
from typing import Optional, Dict, Union
from lazyops.types import BaseSettings, validator, lazyproperty
from async_openai.version import VERSION
from async_openai.types.options import ApiType

_should_reset_api: bool = False

class OpenAISettings(BaseSettings):
    url: Optional[str] = None
    scheme: Optional[str] = 'http://'
    host: Optional[str] = None
    port: Optional[int] = 8080
    
    api_key: Optional[str] = None
    api_path: Optional[str] = '/v1'
    api_type: Optional[ApiType] = ApiType.open_ai
    api_version: Optional[str] = None
    api_key_path: Optional[pathlib.Path] = None
    
    organization: Optional[str] = None
    proxies: Optional[Union[Dict, str]] = None
    app_info: Optional[Dict[str, str]] = None
    debug_enabled: Optional[bool] = False
    ignore_errors: Optional[bool] = False

    timeout: Optional[int] = 600
    max_retries: Optional[int] = 3

    @validator("api_type")
    def validate_api_type(cls, v):
        if v is None: return ApiType.open_ai
        return ApiType(v) if isinstance(v, str) else v
    
    @validator("api_key")
    def validate_api_key(cls, v, values):
        if v is None and values.get('api_key_path') is not None:
            return values['api_key_path'].read_text()
        return v
    

    class Config:
        env_prefix = 'OPENAI_'
        case_sensitive = False

    @lazyproperty
    def api_url(self) -> str:
        if self.url:
            return self.url
        if self.host:
            url = f"{self.scheme}{self.host}"
            if self.port: url += f":{self.port}"
            return url
        
        # Return the official Open API URL
        return "https://api.openai.com"
    

    @lazyproperty
    def base_url(self) -> str:
        if self.api_path:
            from urllib.parse import urljoin
            return urljoin(self.api_url, self.api_path)
        return self.api_url
    
    @lazyproperty
    def base_headers(self) -> Dict[str, str]:
        import platform
        ua = f"OpenAI/v1 async_openai/{VERSION}"
        if self.app_info:
            t = ""
            if "name" in self.app_info:
                t += self.app_info["name"]
            if "version" in self.app_info:
                t += f"/{self.app_info['version']}"
            if "url" in self.app_info:
                t += f" ({self.app_info['url']})"
            ua += f" {t}"
        uname_without_node = " ".join(
            v for k, v in platform.uname()._asdict().items() if k != "node"
        )
        data = {
            "bindings_version": VERSION,
            "httplib": "httpx",
            "lang": "python",
            "lang_version": platform.python_version(),
            "platform": platform.platform(),
            "publisher": "growthengineai",
            "uname": uname_without_node,
        }
        if self.app_info:
            data["application"] = self.app_info
        return {"X-OpenAI-Client-User-Agent": json.dumps(data), "User-Agent": ua}
    
    def get_api_key_headers(
        self,
        api_key: Optional[str] = None, 
        api_type: Optional[Union[ApiType, str]] = None
    ) -> Dict[str, str]:
        if api_key is None: api_key = self.api_key
        if api_type is None: api_type = self.api_type
        api_type = api_type.value if isinstance(api_type, ApiType) else api_type
        if api_type in {"openai", "azure_ad"}:
            return {"Authorization": f"Bearer {api_key}"}
        return {"api-key": api_key}

    @lazyproperty
    def headers(self):
        _headers = self.base_headers.copy()
        if self.api_key: 
            _headers.update(self.get_api_key_headers())
        return _headers

    def get_headers(
        self, 
        api_key: Optional[str] = None, 
        api_version: Optional[str] = None,
        api_type: Optional[Union[ApiType, str]] = None,
        organization: Optional[str] = None,
        **kwargs
    ) -> Dict[str, str]:

        headers = self.headers.copy()
        if kwargs: headers.update(**kwargs)
        headers.update(self.get_api_key_headers(api_key = api_key, api_type = api_type))
        if organization is None: organization = self.organization
        if api_version is None: api_version = self.api_version
        if organization:
            headers["OpenAI-Organization"] = organization
        if api_version is not None and api_type.value == 'open_ai':
            headers["OpenAI-Version"] = api_version
        if self.debug_enabled:
            headers["OpenAI-Debug"] = "true"
        headers['Content-Type'] = 'application/json'
        return headers

    def get_api_url(
        self, 
        host: Optional[str] = None, 
        port: Optional[int] = None, 
        scheme: Optional[str] = None, 
        url: Optional[str] = None,
        **kwargs
    ) -> str:
        if url: return url
        if host:
            url = f"{scheme or self.scheme}{host}"
            if port: url += f":{port}"
            return url
        return self.api_url

    def get_base_api_url(
        self, 
        host: Optional[str] = None, 
        port: Optional[int] = None, 
        scheme: Optional[str] = None, 
        url: Optional[str] = None,
        api_path: Optional[str] = None,
        **kwargs
    ) -> str:
        api_url = self.get_api_url(
            host=host,
            port=port,
            scheme=scheme,
            url=url,
        )
        api_path = api_path or self.api_path
        if api_path:
            from urllib.parse import urljoin
            return urljoin(api_url, api_path)
        return api_url

    def configure(
        self, 
        url: Optional[str] = None,
        scheme: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        
        api_key: Optional[str] = None,
        api_path: Optional[str] = None,
        api_type: Optional[ApiType] = None,
        api_version: Optional[str] = None,
        api_key_path: Optional[pathlib.Path] = None,

        organization: Optional[str] = None,
        proxies: Optional[Union[str, Dict]] = None,
        timeout: Optional[int] = None,
        max_retries: Optional[int] = None,
        app_info: Optional[Dict[str, str]] = None,
        debug_enabled: Optional[bool] = None,
        ignore_errors: Optional[bool] = None,
        **kwargs
    ):
        """
        Allows Post-Init Configuration of the OpenAI Settings

        Usage:

        ```python
        >>> settings.configure(
        >>>    api_key = 'sk-...',
        >>>    organization = 'org-...',
        >>>    max_retries = 4,
        >>>    timeout = 60,
        >>> )
        ```

        :param url: The OpenAI API URL              | Env: [`OPENAI_API_URL`]
        :param scheme: The OpenAI API Scheme        | Env: [`OPENAI_API_SCHEME`]
        :param host: The OpenAI API Host            | Env: [`OPENAI_API_HOST`]
        :param port: The OpenAI API Port            | Env: [`OPENAI_API_PORT`]
        :param api_key: The OpenAI API Key          | Env: [`OPENAI_API_KEY`]
        :param api_path: The OpenAI API Path        | Env: [`OPENAI_API_PATH`]
        :param api_type: The OpenAI API Type        | Env: [`OPENAI_API_TYPE`]
        :param api_version: The OpenAI API Version  | Env: [`OPENAI_API_VERSION`]
        :param api_key_path: The API Key Path       | Env: [`OPENAI_API_KEY_PATH`]
        :param organization: Organization           | Env: [`OPENAI_ORGANIZATION`]
        :param proxies: The OpenAI Proxies          | Env: [`OPENAI_PROXIES`]
        :param timeout: Timeout in Seconds          | Env: [`OPENAI_TIMEOUT`]
        :param max_retries: The OpenAI Max Retries  | Env: [`OPENAI_MAX_RETRIES`]
        :param kwargs: Additional Keyword Arguments

        """
        if url is not None: self.url = url
        if scheme is not None: self.scheme = scheme
        if host is not None: self.host = host
        if port is not None: self.port = port
        if api_key is not None: self.api_key = api_key
        if api_path is not None: self.api_path = api_path
        if api_key_path is not None: 
            self.api_key_path = api_key_path if isinstance(api_key_path, pathlib.Path) else pathlib.Path(api_key_path)
            self.api_key = self.api_key_path.read_text().strip()
        
        if api_type is not None: 
            self.api_type = ApiType(api_type) if isinstance(api_type, str) else api_type
        
        if api_version is not None: 
            self.api_version = self.api_type.get_version(api_version)
        if organization is not None: self.organization = organization
        if proxies is not None: self.proxies = proxies
        if timeout is not None: self.timeout = timeout
        if max_retries is not None: self.max_retries = max_retries
        if app_info is not None: self.app_info = app_info
        if debug_enabled is not None: self.debug_enabled = debug_enabled
        if ignore_errors is not None: self.ignore_errors = ignore_errors



settings = OpenAISettings()


from __future__ import absolute_import

from async_openai.utils.logs import logger
from async_openai.utils.helpers import (
    is_naive,
    total_seconds,
    remove_trailing_slash,
    
)
from async_openai.utils.config import (
    OpenAISettings, 
    settings
)




"""
Client Metaclass
"""

from __future__ import annotations

import pathlib

from typing import Optional, Callable, Dict, Union, TYPE_CHECKING

from async_openai.schemas import *
from async_openai.types.options import ApiType
from async_openai.utils.config import settings

if TYPE_CHECKING:
    from async_openai.client import OpenAIClient


class OpenAIMetaClass(type):
    api_key: Optional[str] = None
    url: Optional[str] = None
    scheme: Optional[str] = None
    host: Optional[str] = None
    port: Optional[int] = None

    api_key: Optional[str] = None
    api_path: Optional[str] = None
    api_type: Optional[ApiType] = None
    api_version: Optional[str] = None
    api_key_path: Optional[pathlib.Path] = None
    
    organization: Optional[str] = None
    proxies: Optional[Union[str, Dict]] = None
    app_info: Optional[Dict[str, str]] = None
    
    debug_enabled: Optional[bool] = None
    ignore_errors: Optional[bool] = None
    max_retries: Optional[int] = None
    on_error: Optional[Callable] = None
    timeout: Optional[int] = None

    headers: Optional[Dict] = None
    _api: Optional['OpenAIClient'] = None

    """
    The Global Meta Class for OpenAI API.
    """


    def configure(
        cls, 
        api_key: Optional[str] = None,
        url: Optional[str] = None,
        scheme: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        
        api_path: Optional[str] = None,
        api_type: Optional[ApiType] = None,
        api_version: Optional[str] = None,
        api_key_path: Optional[pathlib.Path] = None,

        organization: Optional[str] = None,
        proxies: Optional[Union[str, Dict]] = None,
        timeout: Optional[int] = None,
        max_retries: Optional[int] = None,
        app_info: Optional[Dict[str, str]] = None,
        debug_enabled: Optional[bool] = None,
        ignore_errors: Optional[bool] = None,
        on_error: Optional[Callable] = None,

        reset: Optional[bool] = None,
        **kwargs
    ):
        """
        Configure the global OpenAI client.

        :param url: The OpenAI API URL              | Env: [`OPENAI_API_URL`]
        :param scheme: The OpenAI API Scheme        | Env: [`OPENAI_API_SCHEME`]
        :param host: The OpenAI API Host            | Env: [`OPENAI_API_HOST`]
        :param port: The OpenAI API Port            | Env: [`OPENAI_API_PORT`]
        :param api_key: The OpenAI API Key          | Env: [`OPENAI_API_KEY`]
        :param api_path: The OpenAI API Path        | Env: [`OPENAI_API_PATH`]
        :param api_type: The OpenAI API Type        | Env: [`OPENAI_API_TYPE`]
        :param api_version: The OpenAI API Version  | Env: [`OPENAI_API_VERSION`]
        :param api_key_path: The API Key Path       | Env: [`OPENAI_API_KEY_PATH`]
        :param organization: Organization           | Env: [`OPENAI_ORGANIZATION`]
        :param proxies: The OpenAI Proxies          | Env: [`OPENAI_PROXIES`]
        :param timeout: Timeout in Seconds          | Env: [`OPENAI_TIMEOUT`]
        :param max_retries: The OpenAI Max Retries  | Env: [`OPENAI_MAX_RETRIES`]
        :param kwargs: Additional Keyword Arguments
        """
        if url is not None: cls.url = url
        if scheme is not None: cls.scheme = scheme
        if host is not None: cls.host = host
        if port is not None: cls.port = port
        if api_key is not None: cls.api_key = api_key
        if api_path is not None: cls.api_path = api_path
        if api_key_path is not None: 
            cls.api_key_path = api_key_path if isinstance(api_key_path, pathlib.Path) else pathlib.Path(api_key_path)
            cls.api_key = cls.api_key_path.read_text().strip()
        
        if api_type is not None: 
            cls.api_type = ApiType(api_type) if isinstance(api_type, str) else api_type
        
        if api_version is not None: 
            cls.api_version = cls.api_type.get_version(api_version)
        if organization is not None: cls.organization = organization
        if proxies is not None: cls.proxies = proxies
        if timeout is not None: cls.timeout = timeout
        if max_retries is not None: cls.max_retries = max_retries
        if app_info is not None: cls.app_info = app_info
        if debug_enabled is not None: cls.debug_enabled = debug_enabled
        if ignore_errors is not None: cls.ignore_errors = ignore_errors
        if on_error is not None: cls.on_error = on_error

        if reset: cls._api = None
        # if cls._api is None:
        #     cls.get_api(**kwargs)


    def get_api(cls, reset: Optional[bool] = None, **kwargs) -> 'OpenAIClient':
        if cls._api is None or reset:
            from async_openai.client import OpenAIClient
            cls.headers = settings.get_headers(
                api_key = cls.api_key,
                organization = cls.organization,
                api_type = cls.api_type,
                api_version = cls.api_version,
            )
            cls._api = OpenAIClient(
                url = cls.url,
                scheme = cls.scheme,
                host = cls.host,
                port = cls.port,
                api_path = cls.api_path,
                api_type = cls.api_type,
                api_version = cls.api_version,
                organization = cls.organization,
                proxies = cls.proxies,
                timeout = cls.timeout,
                max_retries = cls.max_retries,
                headers = cls.headers,
                app_info = cls.app_info,
                debug_enabled = cls.debug_enabled,
                ignore_errors = cls.ignore_errors,
                on_error = cls.on_error,
                **kwargs
            )
        return cls._api
    

    @property
    def api(cls) -> 'OpenAIClient':
        """
        Returns the inherited OpenAI client.
        """
        if cls._api is None:
            cls.get_api()
        return cls._api
    

    """
    API Routes
    """

    @property
    def completions(cls) -> CompletionRoute:
        """
        Returns the `CompletionRoute` class for interacting with `Completions`.
        
        Doc: `https://beta.openai.com/docs/api-reference/completions`
        """
        return cls.api.completions
    
    @property
    def Completions(cls) -> CompletionRoute:
        """
        Returns the `CompletionRoute` class for interacting with `Completions`.
        
        Doc: `https://beta.openai.com/docs/api-reference/completions`
        """
        return cls.api.completions
    

    @property
    def chat(cls) -> ChatRoute:
        """
        Returns the `ChatRoute` class for interacting with `Chat`.
        
        Doc: `https://beta.openai.com/docs/api-reference/chat`
        """
        return cls.api.chat
    
    @property
    def Chat(cls) -> ChatRoute:
        """
        Returns the `ChatRoute` class for interacting with `Chat`.
        
        Doc: `https://beta.openai.com/docs/api-reference/chat`
        """
        return cls.api.chat
    
    @property
    def edits(cls) -> EditRoute:
        """
        Returns the `EditRoute` class for interacting with `Edits`.
        
        Doc: `https://beta.openai.com/docs/api-reference/edits`
        """
        return cls.api.edits
    
    @property
    def embeddings(cls) -> EmbeddingRoute:
        """
        Returns the `EmbeddingRoute` class for interacting with `Embeddings`.
        
        Doc: `https://beta.openai.com/docs/api-reference/embeddings`
        """
        return cls.api.embeddings
    
    @property
    def images(cls) -> ImageRoute:
        """
        Returns the `ImageRoute` class for interacting with `Images`.
        
        Doc: `https://beta.openai.com/docs/api-reference/images`
        """
        return cls.api.images
    
    @property
    def models(cls) -> ModelRoute:
        """
        Returns the `ModelRoute` class for interacting with `models`.
        
        Doc: `https://beta.openai.com/docs/api-reference/models`
        """
        return cls.api.models

    """
    Context Managers
    """

    async def async_close(cls):
        if cls._api is not None:
            await cls._api.async_close()
    
    def close(cls):
        if cls._api is not None:
            cls._api.close()
    
    def __enter__(cls):
        return cls
    
    def __exit__(cls, exc_type, exc_value, traceback):
        cls.close()
    
    async def __aenter__(cls):
        return cls
    
    async def __aexit__(cls, exc_type, exc_value, traceback):
        await cls.async_close()


from __future__ import absolute_import

from async_openai.utils.config import OpenAISettings, settings
from async_openai.types.errors import (
    OpenAIError,
    APIError,
    AuthenticationError,
    InvalidRequestError,
    RateLimitError,
    APIConnectionError,
    Timeout,
    TryAgain,
    ServiceUnavailableError,
)

from async_openai.types.options import (
    ApiType,
    CompletionModels,
    FilePurpose,
    FinetuneModels,
    ImageSize,
    ImageFormat,
)

## Base Object Models
from async_openai.schemas.completions import CompletionChoice, CompletionObject, CompletionResponse
from async_openai.schemas.chat import ChatMessage, ChatChoice, ChatObject, ChatResponse
from async_openai.schemas.edits import EditChoice, EditObject, EditResponse
from async_openai.schemas.embeddings import EmbeddingData, EmbeddingObject, EmbeddingResponse
# from async_openai.schemas.files import FileChoice, FileObject, FileResponse
from async_openai.schemas.images import ImageData, ImageObject, ImageResponse
from async_openai.schemas.models import ModelData, ModelObject, ModelResponse


## Route Models
from async_openai.schemas.completions import CompletionRoute
from async_openai.schemas.chat import ChatRoute
from async_openai.schemas.edits import EditRoute
from async_openai.schemas.embeddings import EmbeddingRoute
# from async_openai.schemas.files import FileRoute
from async_openai.schemas.images import ImageRoute
from async_openai.schemas.models import ModelRoute



from async_openai.routes import ApiRoutes
from async_openai.client import OpenAIClient, OpenAI



# Completions = OpenAI.completions
# Edits = OpenAI.edits
# Embeddings = OpenAI.embeddings
# # Files = OpenAI.files
# Images = OpenAI.images
# Models = OpenAI.models



import sys
from pathlib import Path
from setuptools import setup, find_packages

if sys.version_info.major != 3:
    raise RuntimeError("This package requires Python 3+")

pkg_name = 'async_openai'
gitrepo = 'GrowthEngineAI/async-openai'

root = Path(__file__).parent
version = root.joinpath('async_openai/version.py').read_text().split('VERSION = ', 1)[-1].strip().replace('-', '').replace("'", '')

requirements = [
    'aiohttpx',
    'file-io',
    'backoff',
]

if sys.version_info.minor < 8:
    requirements.append('typing_extensions')

extras = {}

args = {
    'packages': find_packages(include = [f'{pkg_name}', f'{pkg_name}.*',]),
    'install_requires': requirements,
    'include_package_data': True,
    'long_description': root.joinpath('README.md').read_text(encoding='utf-8'),
    'entry_points': {
        "console_scripts": []
    },
    'extras_require': extras,
}

setup(
    name = pkg_name,
    version = version,
    url=f'https://github.com/{gitrepo}',
    license='MIT Style',
    description='Unofficial Async Python client library for the OpenAI API',
    author='Tri Songz',
    author_email='ts@growthengineai.com',
    long_description_content_type="text/markdown",
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries',
    ],
    **args
)



import datetime
import aiohttpx
import contextlib

from lazyops.types import BaseModel, lazyproperty
from typing import Dict, Optional, Any, List, Type, Union, cast
from async_openai.types.errors import error_handler
from async_openai.types.resources import BaseResource, FileObject, Usage
from async_openai.utils import logger

__all__ = [
    'BaseResponse'
]


"""
Response Class
"""

class BaseResponse(BaseResource):
    
    """
    Base Object class for responses to
    inherit from
    """

    id: Optional[str]
    object: Optional[str]
    created: Optional[datetime.datetime]
    updated: Optional[datetime.datetime]
    
    choices: Optional[List[Type[BaseResource]]]
    data: Optional[Union[Type[BaseResource], List[BaseResource]]] # Used for single retrieval if not in list.
    events: Optional[List[BaseResource]]

    data_model: Optional[Type[BaseResource]] = None
    choice_model: Optional[Type[BaseResource]] = None
    event_model: Optional[Type[BaseResource]] = None

    # Other Metadata
    usage: Optional[Usage] = None
    model: Optional[str] = None

    _input_object: Optional[Type[BaseResource]] = None
    _response: Optional[aiohttpx.Response] = None

    _has_metadata: Optional[bool] = False


    @property
    def has_choices(self):
        return self.choice_model is not None
    
    @property
    def has_data(self):
        return self.data_model is not None
    
    @property
    def has_events(self):
        return self.event_model is not None
    
    @lazyproperty
    def resource_model(self) -> Type[BaseResource]:
        if self.has_choices:
            return self.choice_model
        if self.has_data:
            return self.data_model
        return self.event_model if self.has_events else None

    @lazyproperty
    def excluded_params(self) -> List[str]:
        return [
            "data_model", "choice_model", "event_model", 
            "excluded_params", "input_object", "resource_model", "input_model",
            # "_response", "_has_metadata", "metadata_fields",
        ]
    
    def dict(self, *args, exclude: Any = None, **kwargs):
        """
        Returns the dict representation of the response
        """
        if exclude is None:
            exclude = set()
        exclude = set(exclude) | set(self.excluded_params)
        return super().dict(*args, exclude = exclude, **kwargs)


    """
    Metadata Properties
    """
    @lazyproperty
    def metadata_fields(self) -> List[str]:
        return [
            field.name for field in self.__fields__.values() if \
                field.name not in [
                    "_input_object", "_response", 
                    "data_model",  "choice_model", "event_model",
                    "data", "choices", "events"
                ]
        ]


    @property
    def since_seconds(self) -> int:
        return (datetime.datetime.now(datetime.timezone.utc) - self.created).total_seconds() \
            if self.created else -1
    
    @property
    def headers(self) -> Dict[str, str]:
        return self._response.headers if self._response else {}
    
    @property
    def response_json(self) -> Dict:
        with contextlib.suppress(Exception):
            return self._response.json() if self._response else {}
        return BaseResource.handle_json(self._response.content) if self._response else {}

    @property
    def has_stream(self):
        return "text/event-stream" in self.headers.get("content-type", "")
    
    @property
    def request_id(self) -> Optional[str]:
        return self.headers.get("request-id", self.headers.get("x-request-id"))
    
    @property
    def organization(self) -> str:
        return self.headers.get("openai-organization")

    @property
    def response_ms(self) -> Optional[int]:
        h = self.headers.get("openai-processing-ms")
        return None if h is None else round(float(h))
    
    @property
    def response_timestamp(self) -> Optional[datetime.datetime]:
        dt = self.headers.get("date")
        return None if dt is None else datetime.datetime.strptime(dt, '%a, %d %b %Y %H:%M:%S GMT')

    @property
    def openai_version(self) -> Optional[str]:
        return self.headers.get("openai-version")
    
    @property
    def openai_model(self) -> Optional[str]:
        return self.headers.get("openai-model")
    
    @property
    def model_name(self) -> Optional[str]:
        return self.model.split('-')[1]
    
    """
    Object Data Properties
    """

    @property
    def resource_data(self) -> List[Union[Type[BaseResource], Dict, Any]]:
        """
        Returns the appropriate data object
        """
        if self.has_data: return self.data
        if self.has_choices: return self.choices
        return self.events if self.has_events else []
    

    def __len__(self):
        return len(self.resource_data)
    
    def __iter__(self):
        return iter(self.resource_data)
    
    def __getitem__(self, idx: int) -> Union[Type[BaseModel], Dict, Any]:
        return self.resource_data[idx] if self.resource_data else None
    
    """
    Response Handling Methods
    """

    def handle_metadata(
        self,
        item: Dict,
        **kwargs
    ):
        """
        Sets the appropriate metadata fields
        for the response resource object
        """
        for field in self.metadata_fields:
            if field in item:
                if field == 'usage':
                    self.usage = Usage.parse_obj(item[field])
                    continue
                
                if 'created' in field or 'updated' in field:
                    # Handle created_at and updated_at fields as well
                    key = field.split('_', 1)[0]
                    setattr(self, key, datetime.datetime.fromtimestamp(item[field], datetime.timezone.utc))
                    continue

                if 'files' in field:
                    setattr(self, field, FileObject.create_many(item[field]))
                    continue

                setattr(self, field, cast(type(getattr(self, field)), item[field]))

        self._has_metadata = True

    def handle_data_item(
        self,
        item: Union[Dict, Any],
        **kwargs
    ):
        """
        Handle a single `data` item
        """
        if isinstance(item, dict) and item.get('data'):
            if self.data is None: self.data = []
            data = self.data_model.create_many(item['data'])
            self.data.extend(data)
        
        elif not self.data:
            self.data = self.data_model.parse_obj(item) if self.data_model else item

    
    def handle_choice_item(
        self,
        item: Union[Dict, Any],
        **kwargs
    ):
        """
        Handle a single `choice` item
        """
        if self.choices is None: self.choices = []
        
        if item.get('choices'):
            choices = self.choice_model.create_many(item['choices'])
            self.choices.extend(choices)
        elif self.has_stream:
            self.choices.append(self.choice_model.parse_obj(item))

        
    
    def handle_event_item(
        self,
        item: Union[Dict, Any],
        **kwargs
    ):
        """
        Handle a single `event` item
        """
        if item.get('events'):
            if self.events is None: self.events = []
            events = self.event_model.create_many(item['events'])
            self.events.extend(events)

    def handle_resource_item(
        self, 
        item: Union[Dict, Any],
        **kwargs
    ):
        """
        Handles a single resource item
        and routes it to the appropriate handler
        """
        if not self._has_metadata:
            self.handle_metadata(item, **kwargs)
        
        if self.has_choices and (item.get('choices') or self.has_stream):
            self.handle_choice_item(item, **kwargs)
        
        elif self.has_events and item.get('events'):
            self.handle_event_item(item, **kwargs)
        
        else:
            self.handle_data_item(item, **kwargs)
        
    
    def construct_resource(
        self, 
        **kwargs
    ):
        """
        Constructs the appropriate resource object
        from the response
        """
        if not self.has_stream:
            return self.handle_resource_item(item = self.response_json, **kwargs)
        
        for item in self.handle_stream(
            response = self._response
        ):
            if "error" in item:
                raise error_handler(
                    response = self._response,
                    data = item,
                )
            
            self.handle_resource_item(item = item, **kwargs)


    @classmethod
    def prepare_response(
        cls, 
        response: aiohttpx.Response,
        input_object: Type['BaseResource'],
        **kwargs
    ) -> Type['BaseResponse']:
        """
        Handles the response and returns the appropriate object
        """
        # logger.info(f"Preparing Response for {cls.__name__}")
        resource = cls(
            _response = response,
            _input_object = input_object,
        )
        resource.construct_resource(**kwargs)
        # logger.info(f"Response: {resource}")
        return resource


    def handle_stream_metadata(
        self,
        item: Dict[str, Any],
        **kwargs
    ):
        """
        Handles the stream metadata
        """
        if not self.id and item.get('id'):
            self.id = item['id']
        if not self.created and item.get('created'):
            self.created = datetime.datetime.fromtimestamp(item['created'], datetime.timezone.utc)
        if not self.usage:
            self.usage = Usage(completion_tokens=0)
    
    @lazyproperty
    def choices_results(self) -> List[Type[BaseResource]]:
        """
        Parses the choices first
        """
        if self.choices is None:
            self.construct_resource()
        return self.choices



import aiohttpx

from typing import Optional, Dict, Callable
from async_openai.schemas import *
from async_openai.utils.config import settings

RouteClasses = {
    'completions': CompletionRoute,
    'chat': ChatRoute,
    'edits': EditRoute,
    'embeddings': EmbeddingRoute,
    # 'files': FileRoute,
    'images': ImageRoute,
    'models': ModelRoute,

}

class ApiRoutes:

    """
    Container for all the routes in the API.
    """

    completions: CompletionRoute = None
    chat: ChatRoute = None
    edits: EditRoute = None
    embeddings: EmbeddingRoute = None
    # files: FileRoute = None
    images: ImageRoute = None
    models: ModelRoute = None
    
    def __init__(
        self,
        client: aiohttpx.Client,
        headers: Optional[Dict] = None,
        debug_enabled: Optional[bool] = False,
        on_error: Optional[Callable] = None,
        ignore_errors: Optional[bool] = False,
        timeout: Optional[int] = None,
        max_retries: Optional[int] = None,
        **kwargs
    ):
        self.client = client
        self.headers = headers or settings.get_headers()
        self.debug_enabled = debug_enabled
        self.on_error = on_error
        self.ignore_errors = ignore_errors
        self.timeout = timeout
        self.max_retries = max_retries
        self.kwargs = kwargs or {}
        self.init_routes()
    

    def init_routes(self):
        for route, route_class in RouteClasses.items():
            setattr(self, route, route_class(
                client = self.client,
                headers = self.headers,
                debug_enabled = self.debug_enabled,
                on_error = self.on_error,
                ignore_errors = self.ignore_errors,
                timeout = self.timeout,
                max_retries = self.max_retries,
                **self.kwargs
            ))
    
